%function Y = imagequilt4(X, tilesize, n, overlap, err)
%Performs the Efros/Freeman Image quilting algorithm on the input
%
%Inputs
%   X:  The source image to be used in synthesis
%   tilesize:   the dimensions of each square tile.  Should divide size(X) evenly
%   n:  The number of tiles to be placed in the output image, in each dimension
%   overlap: The amount of overlap to allow between pixels (def: 1/6 tilesize)
%   err: used when computing list of compatible tiles (def: 0.1)

function [Y, subInfo, cutInfo] = imagequilt4(X, tilesize, n_row, n_col, overlap, err)

X = double(X);

if( length(size(X)) == 2 )
    X = repmat(X, [1 1 3]);
elseif( length(size(X)) ~= 3 )
    error('Input image must be 2 or 3 dimensional');
end;

if( nargin < 6 )
    err = 0.05;
end;

if( nargin < 5 )
    overlap = round(tilesize / 6);
end;

if( overlap >= tilesize )
    error('Overlap must be less than tilesize');
end;

destsize_row = n_row * tilesize - (n_row-1) * overlap; 
destsize_col = n_col * tilesize - (n_col-1) * overlap ;

X2 = X.^2;

Y = zeros(destsize_row, destsize_col, size(X,3));

subInfo = zeros(n_row,n_col,2);
cutInfo = zeros(n_row,n_col,tilesize*tilesize);

for i=1:n_row,
for j=1:n_col,
    
	startI = (i-1)*tilesize - (i-1) * overlap + 1;
	startJ = (j-1)*tilesize - (j-1) * overlap + 1;
	endI = startI + tilesize -1 ;
	endJ = startJ + tilesize -1;
     
	%Determine the distances from each tile to the overlap region
	%This will eventually be replaced with convolutions
	distances = zeros( size(X,1)-tilesize, size(X,2)-tilesize );         
                             
    %Compute the distances from the source to the left overlap region
    if( j > 1 )
        distances = ssd4( X, X2, Y(startI:endI, startJ:startJ+overlap-1, :) );    
        distances = distances(1:end, 1:end-tilesize+overlap);
    end;
    
    %Compute the distance from the source to top overlap region
    if( i > 1 )
        Z = ssd4( X, X2, Y(startI:startI+overlap-1, startJ:endJ, :) );
        Z = Z(1:end-tilesize+overlap, 1:end);
        if( j > 1 ) 
            distances = distances + Z;
        else 
            distances = Z;
        end;
    end;
    
    %If both are greater, compute the distance of the overlap
    if( i > 1 && j > 1 )
        Z = ssd4( X, X2, Y(startI:startI+overlap-1, startJ:startJ+overlap-1, :) );
        Z = Z(1:end-tilesize+overlap, 1:end-tilesize+overlap);                   
        distances = distances - Z;
    end;
                    
	%Find the best candidates for the match
    t_idx = (distances>=0);
	best = min(distances(t_idx));
	candidates = find(distances(:) <= (1+err)*best);
	
	idx = candidates(ceil(rand(1)*length(candidates)));
                 
	[sub(1), sub(2)] = ind2sub(size(distances), idx);
	fprintf( '[%02d %02d / %02d %02d] Picked tile (%d, %d) out of %d candidates.  Best error=%.4f\n', i, j, n_row, n_col, sub(1), sub(2), length(candidates), best );       
    
    subInfo(i,j,1) = sub(1);
    subInfo(i,j,2) = sub(2);
             
	%Initialize the mask to all ones
	M = ones(tilesize, tilesize);
	
	%We have a left overlap
	if( j > 1 )
     
		%Compute the SSD in the border region
 		E = ( X(sub(1):sub(1)+tilesize-1, sub(2):sub(2)+overlap-1) - Y(startI:endI, startJ:startJ+overlap-1) ).^2;
		%E = sum(( X(sub(1):sub(1)+tilesize-1, sub(2):sub(2)+overlap-1, :) - Y(startI:endI, startJ:startJ+overlap-1, :) ).^2, 3);
        
		%Compute the mincut array
		C = mincut(E, 0);
		
		%Compute the mask and write to the destination
		M(1:end, 1:overlap) = double(C >= 0);
		    
	end;
	
	%We have a top overlap
	if( i > 1 )
  
         %Compute the SSD in the border region
         E = ( X(sub(1):sub(1)+overlap-1, sub(2):sub(2)+tilesize-1) - Y(startI:startI+overlap-1, startJ:endJ) ).^2;
         %E = sum(( X(sub(1):sub(1)+overlap-1, sub(2):sub(2)+tilesize-1,:) - Y(startI:startI+overlap-1, startJ:endJ,:) ).^2, 3);
        
         %Compute the mincut array
         C = mincut(E, 1);
         
         %Compute the mask and write to the destination
         M(1:overlap, 1:end) = M(1:overlap, 1:end) .* double(C >= 0);
    
	end;
              
	if( i == 1 && j == 1 )
		Y(startI:endI, startJ:endJ, :) = X(sub(1):sub(1)+tilesize-1, sub(2):sub(2)+tilesize-1, :);
	else
		%Write to the destination using the mask
		Y(startI:endI, startJ:endJ, :) = filtered_write(Y(startI:endI, startJ:endJ, :), ...
						X(sub(1):sub(1)+tilesize-1, sub(2):sub(2)+tilesize-1, :), M); 
	end;
 
    cutInfo(i,j,:) = M(:);
end;
end;

% figure;
% image(uint8(Y));


%---------------------------------------------------------------------------
function A = filtered_write(A, B, M)

MA = (M==0);
MB = (M==1);
for i = 1:size(B,3),
    %A(:, :, i) = A(:,:,i) .* (M == 0) + B(:,:,i) .* (M == 1);
    a = A(:,:,i);
    b = B(:,:,i);
    a(MB) = b(MB);
    A(:,:,i) = a;
end;
% 
% height = size(A,1);
% width = size(A,2);
% N = edge(M);
% idx = find(N==1);
% [subi,subj] = ind2sub([height,width],idx);
% 
% for i = 1:length(idx)
%     A(subi(i), subj(i), 1) = 0;
%     A(subi(i), subj(i), 2) = 0;
%     A(subi(i), subj(i), 3) = 255;
% end

%function Z = ssd2(X, Y)
%Computes the sum of squared distances between X and Y for each possible
% overlap of Y on X.  Y is thus smaller than X
%
%Inputs:
%   X - larger image
%   Y - smaller image
%
%Outputs:
%   Each pixel of Z contains the ssd for Y overlaid on X at that pixel

function Z = ssd4(X, X2, Y)

K = ones(size(Y,1), size(Y,2));

Z = 0;
%for k=1:size(X,3),
for k = 1:1
    A = X(:,:,k);
    A2 = X2(:,:,k);
    B = Y(:,:,k);
    
    a2 = filter2(K, A2, 'valid');
    b2 = sum(sum(B.^2));
    ab = filter2(B, A, 'valid').*2;

    Z = Z + ((a2 - ab) + b2);
end;

